__title__ = 'algoseek-connector'
__description__ = 'A wrapper library for ORM-like SQL builder and executor'
__version__ = '1.0.2'
__author__ = 'Taras Kuzyo'
__author_email__ = 'taras@algoseek.com'
__copyright__ = 'Copyright 2022 algoseek, LLC'
__license__ = "Apache 2.0"
